hidden_alphabet = '\u200B\u200C\u200D\u202F'
hidden_alphabet_table = {
    '\u200B': 0,
    '\u200C': 1,
    '\u200D': 2,
    '\u202F': 3,
}


def encode_byte(b):
    r = ''
    for _ in range(4):
        r = r + hidden_alphabet[b % 4]
        b //= 4
    return r


def decode_hidden_byte(h: str):
    b = 0
    for i, hb in enumerate(h):
        b |= (hidden_alphabet_table[hb] << i * 2)
    return b


def encode(ss: str):
    r = ''
    for b in ss.encode('ascii'):
        r = r + encode_byte(b)
    return r


def decode(ss: str):
    r = bytearray()
    for b in range(len(ss) // 4):
        r.append(decode_hidden_byte(ss[b * 4:b * 4 + 4]))
    return r.decode('ascii')



from __future__ import absolute_import

import itertools
from typing import (
    Iterable,
    List,
)


def chunks(list: List, n: int):
    for i in range(0, len(list), n):
        yield list[i:i + n]


def ichunks(iterable: Iterable, n: int):
    it = iter(iterable)
    while True:
        chunk = tuple(itertools.islice(it, n))
        if not chunk:
            return
        yield chunk



import unittest

from izihawa_utils.coding import (
    base58_check_decode,
    base58_check_encode,
    base58_decode,
    base58_encode,
    base256_decode,
    base256_encode,
    count_leading_chars,
)


class TestCoding(unittest.TestCase):
    def test_count_leading_characters(self):
        self.assertEqual(count_leading_chars('a\0bcd\0', '\0'), 0)
        self.assertEqual(count_leading_chars('\0\0a\0bcd\0', '\0'), 2)
        self.assertEqual(count_leading_chars('1a\0bcd\0', '1'), 1)

    def test_base256(self):
        self.assertEqual(base256_encode(base256_decode(b'abc')), b'abc')
        self.assertEqual(base256_encode(0x4142), b'AB')
        self.assertEqual(base256_decode(b'AB'), 0x4142)

    def test_base58(self):
        self.assertEqual(base58_encode(base58_decode('abc')), 'abc')
        self.assertEqual(base58_decode('121'), 58)
        self.assertEqual(base58_decode('5HueCGU8rMjxEXxiPuD5BDku4MkFqeZyd4dZ1jvhTVqvbTLvyTJ'),
                         0x800C28FCA386C7A227600B2FE50B7CAE11EC86D3BF1FBE471BE89827E19D72AA1D507A5B8D)

    def test_base58_check(self):
        self.assertEqual(base58_check_decode(base58_check_encode(b'abc', 42)), (b'abc', 42))
        self.assertEqual(base58_check_decode(base58_check_encode(b'\0\0abc', 0)), (b'\0\0abc', 0))
        s = base256_encode(0x0C28FCA386C7A227600B2FE50B7CAE11EC86D3BF1FBE471BE89827E19D72AA1D)
        b = base58_check_encode(s, 0x80)
        self.assertEqual(b, "5HueCGU8rMjxEXxiPuD5BDku4MkFqeZyd4dZ1jvhTVqvbTLvyTJ")



from izihawa_utils.text import camel_to_snake


def test_camel_to_snake():
    assert camel_to_snake('CamelCase') == 'camel_case'
    assert camel_to_snake('camelCase') == 'camel_case'
    assert camel_to_snake('camelCase camel123Case') == 'camel_case camel123_case'
    assert camel_to_snake('camelCase\ncamelCase') == 'camel_case\ncamel_case'



def import_object(name):
    components = name.split('.')
    mod = __import__(components[0])
    parts = [components[0]]
    for comp in components[1:]:
        parts.append(comp)
        try:
            mod = __import__('.'.join(parts), fromlist=[''])
        except (ImportError, ModuleNotFoundError):
            mod = getattr(mod, comp)
    return mod


def instantiate_object(descriptor):
    return import_object(descriptor['class'])(**descriptor.get('kwargs', {}))



import errno
import hashlib
import os
import re
from grp import getgrnam
from pwd import getpwnam


def tail(f, lines=20):
    readlines = f.readlines()
    return ''.join(readlines[-lines:])


def grep(f, patter):
    result = []
    for line in f:
        if re.search(patter, line):
            result.append(line)
    return result


def mkdir_p(filepath):
    try:
        os.makedirs(filepath)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(filepath):
            pass
        else:
            raise


class chdir(object):
    def __init__(self, directory):
        self.directory = directory
        self.current_dir = os.getcwd()

    def __enter__(self):
        os.chdir(self.directory)

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self.current_dir)


def parse_owner(owner):
    owner = owner or ''
    owner = owner.strip()

    user = None
    group = None

    splitted = owner.split(':', 2)
    if len(splitted) > 0:
        user = splitted[0].strip() or None
        if len(splitted) > 1:
            group = splitted[1].strip() or None

    return user, group


def get_uid(name):
    """Returns an uid, given a user name."""
    if getpwnam is None or name is None:
        return None
    try:
        result = getpwnam(name)
    except KeyError:
        result = None
    if result is not None:
        return result[2]
    return None


def get_gid(name):
    """Returns a gid, given a group name."""
    if getgrnam is None or name is None:
        return None
    try:
        result = getgrnam(name)
    except KeyError:
        result = None
    if result is not None:
        return result[2]
    return None


def parent_chain(filepath):
    filepath = os.path.abspath(filepath)
    while filepath != '/':
        filepath = os.path.dirname(filepath)
        yield filepath


def md5(file):
    hash_md5 = hashlib.md5()
    for chunk in iter(lambda: file.read(4096), b""):
        hash_md5.update(chunk)
    return hash_md5.hexdigest()



import logging

from izihawa_utils.tskv import TskvRecord


class BaseError(Exception):
    code = 'base_error'
    level = logging.ERROR
    info = dict()
    internal_info = dict()

    def __init__(self, **kwargs):
        self.info = dict(self.info)
        self.internal_info = dict(self.internal_info)

        if 'code' in kwargs:
            self.code = kwargs['code']

        if 'internal_info' in kwargs:
            self.internal_info.update(kwargs.pop('internal_info'))

        self.info['code'] = self.code
        self.info.update(kwargs)

    def as_dict(self):
        return dict(self.info)

    def as_tskv(self):
        return TskvRecord(self.info)

    def as_internal_dict(self):
        info = self.as_dict()
        info.update(self.internal_info)
        return info

    def as_internal_tskv(self):
        info = self.as_tskv()
        info.update(self.internal_info)
        return info

    def __repr__(self):
        r = self.as_internal_tskv()
        return str(r)

    def __str__(self):
        return repr(self)

    def add(self, key, value):
        self.info[key] = value


class MissingFileError(BaseError, FileNotFoundError):
    def __init__(self, file_path=''):
        self.info = {'path': file_path}


class NeedRetryError(BaseError):
    pass



import enum
import types
from datetime import (
    date,
    datetime,
)

from izihawa_utils.common import filter_none


class JsonSerializator:
    def __init__(self, max_serialization_depth, custom_field_processors=None, custom_type_processors=None):
        self._max_serialization_depth = max_serialization_depth
        self._custom_field_processors = custom_field_processors or {}
        self._custom_type_processors = custom_type_processors or {}

    def _process_list(self, field, depth, excludes=None, includes=None, *args, **kwargs):
        depth += 1
        if depth > self._max_serialization_depth:
            return

        r = [self.process_field(x, depth, excludes=excludes, includes=includes, *args, **kwargs) for x in field]
        return filter_none(r)

    def _process_dict(self, field, depth, excludes=None, includes=None, *args, **kwargs):
        depth += 1
        result = {}
        if depth > self._max_serialization_depth:
            return result
        for key in field:
            if excludes and key in excludes:
                continue
            if key in self._custom_field_processors:
                value = self._custom_field_processors[key](
                    field[key], depth, excludes=excludes, includes=includes, *args, **kwargs
                )
            else:
                value = self.process_field(field[key], depth, excludes=excludes, includes=includes, *args, **kwargs)
            if value is None:
                continue
            key = self.process_field(key, depth, excludes=excludes, includes=includes, *args, **kwargs)
            result[key] = value
        return result

    def process_field(self, field, depth, excludes=None, includes=None, *args, **kwargs):
        for type_, processor in self._custom_type_processors.items():
            if isinstance(field, type_):
                return processor(field, depth, excludes=excludes, includes=includes, *args, **kwargs)
        if isinstance(field, (date, datetime)):
            return field.strftime('%Y-%m-%d %H:%M:%S')
        if isinstance(field, (enum.Enum, enum.IntEnum)):
            return field.name
        if isinstance(field, (list, tuple, set)) or isinstance(field, types.GeneratorType):
            return self._process_list(field, depth, excludes=excludes, includes=includes, *args, **kwargs)
        if isinstance(field, dict):
            return self._process_dict(field, depth, excludes=excludes, includes=includes, *args, **kwargs)
        return field


class JsonSerializable:
    serializator = JsonSerializator
    max_serialization_depth = 1

    def serialize(self, max_depth=None, excludes=None, includes=None, *args, **kwargs):
        serializator = self.serializator(max_serialization_depth=max_depth or self.max_serialization_depth)
        return serializator.process_field(self, depth=0, excludes=excludes, includes=includes, *args, **kwargs)



MAX_LOG_LENGTH = 8192
REPLACE_MAP = [
    ('\\', r'\\'),
    ('\t', r'\t'),
    ('\n', r'\n'),
    ('\r', r'\r'),
    ('\0', r'\0'),
]


class TskvRecord(dict):
    preferred_fields = ['unixtime', 'datetime', 'request_id', 'user_ip', 'path']

    @staticmethod
    def escape_value(text):
        for char, replacement in REPLACE_MAP:
            text = text.replace(char, replacement)
        return text

    @staticmethod
    def truncate_str(text):
        return text[:MAX_LOG_LENGTH]

    def __str__(self):
        elements = []
        for preferred_field in self.preferred_fields:
            if self.get(preferred_field) is not None:
                elements.append(self.escape_value('%s=%s' % (preferred_field, self[preferred_field])))
                self.pop(preferred_field)
        for k, v in sorted(self.items()):
            if v is not None:
                elements.append(self.escape_value('%s=%s' % (k, v)))
        return self.truncate_str('\t'.join(elements))
